/* Screenshot flash */
@keyframes flash_animation {
    50% {
        opacity: 0.2;
    }
}

.flash {
    border: 1px solid transparent;
    animation: flash_animation;
    animation-duration: 300ms;
    animation-iteration-count: 1;
}

.thumbnail {
    background: @theme_bg_color;
    color: @theme_fg_color;
    border: 1px solid @borders;
}

/* Onboarding dialog */
.onboarding-dialog-page {
    background-repeat: no-repeat;
    background-size: 60%;
    background-position: center 15%;
}
